//*************************************************
// File: main.cpp
//
// Converter of MPASM code into SASM
//
// MPASM code may be embedded into an HTM file
// (In this case several symbols, like >, < , are
// converted to HTML equivalents).
//
// Command line:
// mpasm2sasm {-s} {-m} input_file_name
//
// -s option tells the converter to include commented source in the output file
// -m option makes lower case mnemonics
// -h  help screen
// -rx Default radix
//     x = 10 - decimal
//     x = 16 - hexadecimal
//  -c  Convert constants in comments
//
// Sep 22, 2000
//		1) Fixed gt; problem of missing ; when in command line
//		for HTM files
//		2) Fixed retw #xx problem of redundant #
//		3) Added recognition of status flags if named "carry",
//		"digitcarry", "zero"
//		4) Added MPASM macros (almost all now are supported)
//		5) Added conversion of CBLOCK!
//
//
// (c) 2000 by Nikolai Golovchenko
//*************************************************

//#include <stdio.h>
//#include <stdlib.h>
//#include<string.h>

//Standard C++ library:
#include <string>
using namespace std;
//include CGI library headers
extern "C"
{
	#include "../../CodeGenerators/cgihtml/cgi-lib.h"
	#include "../../CodeGenerators/cgihtml/html-lib.h"
	char** CommandLineArguments;
}
#include "MPASMFileLine.h"
#include "../../CodeGenerators/HTML/HTML.h"
#include "HTML_PIC2SXPage.h"
#include "HTML_PIC2SXPage2.h"
#include "main.h"

//Global variables
bool isCGI;		//is in CGI mode? (handled by main)

int
main(int argc, char *argv[ ], char *envp[ ])
{
	FILE* fin;
	FILE* fout;
	string inName;
	string outName;
	MPASMFileLine x;
	llist entries = {NULL}; //clear list
	HTML_PIC2SXPage* page;
	char* value;

	/********* DEBUG *****************
	_putenv("SERVER_SOFTWARE=Microsoft-IIS/4.0");
	_putenv("SERVER_NAME=www.piclist.com");
	_putenv("GATEWAY_INTERFACE=CGI/1.1");
	_putenv("SERVER_PROTOCOL=HTTP/1.1");
	_putenv("SERVER_PORT=80");
	_putenv("REQUEST_METHOD=GET");
	_putenv("PATH_TRANSLATED=C:\\Inetpub\\wwwroot");
	_putenv("SCRIPT_NAME=/cgi-bin/test.exe");
	_putenv("REMOTE_HOST=204.210.50.240");
	_putenv("REMOTE_ADDR=204.210.50.240");
	_putenv("CONTENT_LENGTH=0");
	_putenv("QUERY_STRING=mpasm=0%0D%0A1%0D%0A2%0D%0A3%0D%0A4&lowercase=on&radix=dec&convert=Convert+to+SX+assembly%21");
	/***********************************/
	
	//pass comand line to CGI module
	CommandLineArguments = argv;

	isCGI = getenv("REQUEST_METHOD") != NULL;
	if(!isCGI)
	{
		//check command line
		argv = checkCommandLine(argc, argv, envp);
		if(argv == 0)
		{
			return(0);
		}
		inName = argv[0];

		//open input file
		fin = fopen(inName.c_str(), "r");
		if(fin == NULL)
		{
			error("Can't open the input file");
			return(2);
		}
		if(inName.find(".htm") != -1)
		{
			//Indicate we are working with HTM file
			MPASMFileLine::isHTML = true;
		}

		//open output file
		newOutName(inName, outName);
		fout = fopen(outName.c_str(), "w");
		if(fout == NULL)
		{
			error("Can't open the output file");
			return(3);
		}

		//read file
		while(!feof(fin))
		{
			x.initialize();
			x.readSourceLine(fin);
			x.extractMnemonic();
			x.translateLineTo(fout);
		}
		//close output file
		fclose(fout);
		//close input file
		fclose(fin);
		return(0);
	}

	// Read form values
	read_cgi_input(&entries);

	//print HTML header for server
	html_header(); 

	//check if form is empty
	if(!is_field_exists(entries, "convert"))
	{
		//first page
		page = new HTML_PIC2SXPage();
	}
	else
	{
		//secon page - results
		page = new HTML_PIC2SXPage2();
	}
	//pass data from form to page object
	value= cgi_val(entries, "mpasm");
	if(value != NULL)
	{
		page->setInput(value);
	}
	value= cgi_val(entries, "lowercase");
	if(value != NULL && strcmp(value, "on") == 0)
	{
		MPASMFileLine::lowerCase = true;
	}
	value= cgi_val(entries, "echo");
	if(value != NULL && strcmp(value, "on") == 0)
	{
		MPASMFileLine::echoSourceLine = true; 
	}
	value= cgi_val(entries, "comments");
	if(value != NULL && strcmp(value, "on") == 0)
	{
		MPASMFileLine::fixConstantsInComments = true;
	}
	value= cgi_val(entries, "radix");
	if(value != NULL)
	{
		if(strcmp(value, "dec") == 0)
		{
			MPASMFileLine::radix = 10;
		}
		else if(strcmp(value, "hex") == 0)
		{
			MPASMFileLine::radix = 16;
		}
	}

	page->generatePage();
	page->print();
	delete(page);
	list_clear(&entries);
	return(0);
}

char**
checkCommandLine(int argc, char *argv[ ], char *envp[ ])
{
	while(*++argv)
	{
		if(argv[0][0] == '-')
		{
			switch(argv[0][1])
			{
			case 's':
				//Echo each source line
				MPASMFileLine::echoSourceLine = true; 
				break;
			case 'm':
				MPASMFileLine::lowerCase = true;
				break;
			case 'h':
				helpScreen();
				break;
			case 'r':
				if(argv[0][2] == '1' && argv[0][3] == '0')
				{
					MPASMFileLine::radix = 10;
				}
				if(argv[0][2] == '1' && argv[0][3] == '6')
				{
					MPASMFileLine::radix = 16;
				}
				break;
			case 'c':
				MPASMFileLine::fixConstantsInComments = true;
				break;
			default:
				error("Unknown option");
				return(0);
				break;
			}
		}
		else
		{
			break;
		}
	}
	if(*argv == NULL)
	{
		helpScreen();
		return(0);
	}
	return(argv);
}

void
error(char* message)
{
	printf("ERROR> %s\n", message);
}

void
newOutName(string& inName, string& outName)
{
	int i;

	//copy inName to outName
	outName = inName;
	//find last '.'
	i = outName.find_last_of(".");
	//insert "_sx" before it
	if(i == -1)
	{
		outName.append("_sx");
	}
	else
	{
		outName.insert(i, "_sx");
	}
}

void
helpScreen()
{
	static bool printed = false;
	if(!printed)
	{
		printed = true;
		printf("------------------------------------------------------\n");
		printf("ASSEMBLER CONVERTER FROM MPASM TO SASM/SX-KEY\n");
		printf("(c) 2000 by Nikolai Golovchenko\n");
		printf("------------------------------------------------------\n");
		printf("Command line: mpasm2sasm.exe {options} source_file\n");
		printf("Options:\n\t-m\tGenerate lower case mnemonics\n");
		printf("\t-rn\tDefault radix in source\n\t\tn = 10\tdecimal\n\t\tn = 16\thexadecimal\n");
		printf("\t-s\tKeep commented source lines\n");
		printf("\t-c\tConvert constants in comments\n");
		printf("\t-h\tHelp (this screen)\n");
	}
}
