//**************************************************
//
// MPASMFileLine class member functions definition
//
//**************************************************

#include "stdio.h"
#include "string"
using namespace std;

#include "MPASMFileLine.h"

//Iniitialize static member
bool MPASMFileLine::isHTML = false;
bool MPASMFileLine::echoSourceLine = false;
bool MPASMFileLine::lowerCase = false;
char MPASMFileLine::radix = 10;
bool MPASMFileLine::fixConstantsInComments = false;

MPASMFileLine::MPASMFileLine()
{
	this->initialize();
	inCblock = false;	//special initialization
	//because cblock action spans over many lines
}
MPASMFileLine::~MPASMFileLine()
{
}

void
MPASMFileLine::initialize()
{
	//label = "";
	//whitespaces = "";
	//alphaField = "";
	sourceLine = "";
	sourceField = "";
	mnemonicStart = 0;
	mnemonic = "";
	mnemonic2 = "";
	operand1 = "";
	operand2 = "";
	comment = "";
	warning = "";
	operandsType = NotApplicable;
}

void
MPASMFileLine::readSourceLine(FILE* fin)
{
	char buffer[256] = "";
	
	fgets(buffer, 256, fin);
	sourceLine.append(buffer);
}

void
MPASMFileLine::readSourceLine(string& fin, int& start)
{
	string buf;
	int end;
	
	end = fin.find("\x0A", start);
	if(end != -1)
	{
		sourceLine.append(fin.substr(start, end - start));
		start = end + 1;
	}
	else
	{
		sourceLine.append(fin.substr(start));
		start = end;
	}
}

void
MPASMFileLine::extractMnemonic()
{
	char ch;
	int x = -1;
	int s = 0;
	bool wasLabel = false;
	bool wasSpaceBeforeLabel = false;
	bool wasSpaceAfterLabel = false;
	bool labelMayBeMnemonic = true;
	bool firstTry = true;
	
	if(inCblock)
	{
		extractCblock();
		return;
	}
	sourceLine.append(" "); //add a guaranteed space
	while(s < sourceLine.size())
	{
		ch = sourceLine.at(s);
		if(isspace(ch))
		{
			//whitespace
			if(!wasSpaceBeforeLabel && !wasLabel)
			{
				wasSpaceBeforeLabel = true;
			}
			if(!wasSpaceAfterLabel && wasLabel)
			{
				wasSpaceAfterLabel = true;
				//end of label - check if this is a mnemonic
				sourceField = sourceLine.substr(x, s - x);
				if(labelMayBeMnemonic)
				{
					decodeSourceLine();
				}
				if(operandsType == NotApplicable)
				{
					//no, it's a label - repeat once more
					if(firstTry)
					{
						wasSpaceBeforeLabel = true;
						wasLabel = false;
						wasSpaceAfterLabel = false;
						firstTry = false;
						labelMayBeMnemonic = true;
					}
					else
					{
						//yes, remember position and quit
						mnemonicStart = s;
						break;
					}
				}
				else
				{
					//yes, it is a mnemonic
					mnemonicStart = x;
					break;
				}
			}
		}
		else if(isalnum(ch)
			|| ch == '<'
			|| ch == '_'
			|| ch == '>'
			|| ch == ':')
		{
			//any character accepted as a 'label': a-z, A-Z, 0-9, <, >, _, :
			//this includes HTML tags
			if(!wasLabel)
			{
				wasLabel = true;
				x = s; //remember start of label
			}
			if(!isalpha(ch))
			{
				labelMayBeMnemonic = false;
			}
		}
		s++;
	}
	//remove space
	sourceLine.erase(sourceLine.size() - 1, 1);
	
	if(operandsType != NotApplicable)
	{
		extractOperands(s);
	}
}

void
MPASMFileLine::extractCblock()
{
	string buf;
	string reg;
	string reg_size;
	int x;
	int s;

	//replace all registers from sourceLine
	//by DS directive to mnemonic
	//If ENDC is encountered, switch mode

	//copy line to buf
	buf = sourceLine;
	//extract comment
	x = 0;
	while(x != -1)
	{
		x = buf.find(';', x);
		//check if this is a part of special charatcters:
		//&amp;&lt;&gt;&quot;
		if(x != -1
			&& buf.substr(x >=4 ? x - 4 : 0, 4) != "&amp"
			&& buf.substr(x >=3 ? x - 3 : 0, 3) != "&lt"
			&& buf.substr(x >=3 ? x - 3 : 0, 3) != "&gt"
			&& buf.substr(x >=5 ? x - 5 : 0, 5) != "&quot")
		{
			//no, it's comment
			comment = buf.substr(x, buf.size() - x);
			stripBlanks(comment);
			buf.resize(x);
		}
		 if(x != -1)
		 {
			 x++;
		 }
	}
	//check for ENDC
	mnemonic = buf;
	operandsType = Directive;
	uppercaseString(mnemonic);
	x = mnemonic.find("ENDC");
	if(x != -1)
	{
		mnemonic=""; //erase ENDC
		inCblock = false; //switch mode
		return;
	}
	
	//no ENDC, just regs or nothing
	mnemonic = "";
	//take a look for a register terminated by a comma
	x = 0;
	while(x != -1)
	{
		//copy rest of buffer to reg
		reg = buf;

		x = buf.find(',');
		//work with current reg
		if(x != -1)
		{
			buf.erase(0, x+1); //reject one reg and comma from buf
			reg.resize(x); //limit reg and exclude comma
		}
		//find optional size specifier ':1'
		s = reg.find(':');
		if(s != -1)
		{
			reg_size = reg.substr(s+1, reg.size() - s-1);
			reg.resize(s); //limit reg and exclude ':'
		}
		else
		{
			reg_size = "1"; //default size
		}
		//strip blanks
		stripBlanks(reg);
		stripBlanks(reg_size);
		//include DS directive
		s = reg.size();
		if(s != 0)
		{
			mnemonic.append(reg);
			do
			{
				mnemonic.append(" ");
				s++;
			} while(s < 16);

			mnemonic.append("DS\t");
			mnemonic.append(reg_size);
			if(x != -1)
			{
				//add enter for intermediate regs
				mnemonic.append("\n");
			}
		}

	}
}
void
MPASMFileLine::decodeSourceLine()
{
	char ch[2] = "";
	
	//copy alphaField to mnemonic and convert to upper case
	mnemonic = sourceField;
	uppercaseString(mnemonic);
	//translate mnemonics
	if(mnemonic == "")
	{
		return;
	}
	//PIC12 instructions (compatible one for one)
	table("ANDLW", "AND\tW, ", Literal);
	table("CALL", "CALL", Label);
	table("CLRWDT", "CLR\t!WDT", NoOperand);        //do nothing more
	table("GOTO", "JMP", Label);
	table("IORLW", "OR\tW, ", Literal);
	table("MOVLW", "MOV\tW, ", Literal);
	table("OPTION", "MOV\t!OPTION, W", NoOperand);
	table("RETLW", "RETW\t", LiteralNoSharp); //literal, but without #
	table("SLEEP", "SLEEP", NoOperand);
	warn("TRIS", "TRIS expanded in two instructions. Check if previous instruction is a skip instruction.");
	table("TRIS", "MOV\tM, ", "MOV\t!R, W", PortX);
	table("XORLW", "XOR\tW, ", Literal);
	table("ADDWF", "ADD", FileAndDestination);
	table("ANDWF", "AND", FileAndDestination);
	table("CLRF", "CLR\tf", FileOperand);   
	table("CLRW", "CLR\tW", NoOperand);
	table("COMF", "NOT\tf", "MOV\tW, /f", TwoChoices);
	table("DECF", "DEC\tf", "MOV\tW, --f", TwoChoices);
	table("DECFSZ", "DECSZ\tf", "MOVSZ\tW, --f", TwoChoices);
	table("INCF", "INC\tf", "MOV\tW, ++f", TwoChoices);
	table("INCFSZ", "INCSZ\tf", "MOVSZ\tW, ++f", TwoChoices);
	table("IORWF", "OR", FileAndDestination);
	table("MOVF", "TEST\tf", "MOV\tW, f", TwoChoices);
	table("MOVWF", "MOV\tf, W", FileOperand);
	table("NOP", "NOP", NoOperand);
	table("SUBWF", "SUB\tf, W", "MOV\tW, f-w", TwoChoices);
	if(isHTML)
	{
		table("RLF", "RL\tf", "MOV\tW, &lt;&lt;f", TwoChoices);
		table("RRF", "RR\tf", "MOV\tW, &gt;&gt;f", TwoChoices);
		table("SWAPF", "SWAP\tf", "MOV\tW, &lt;&gt;f", TwoChoices);
	}
	else
	{
		table("RLF", "RL\tf", "MOV\tW, <<f", TwoChoices);
		table("RRF", "RR\tf", "MOV\tW, >>f", TwoChoices);
		table("SWAPF", "SWAP\tf", "MOV\tW, <>f", TwoChoices);
	}
	table("XORWF", "XOR", FileAndDestination);
	table("BCF", "CLRB", BitOperand);
	table("BSF", "SETB", BitOperand);
	table("BTFSC", "SNB", BitOperand);
	table("BTFSS", "SB", BitOperand);
	//PIC16 instructions (not all compatible)
	warn("ADDLW", "ADDLW was expanded in three instructions! Check if previous instruction is a skip instruction. ");
	warn("SUBLW", "Manual replacement required for \"SUBLW k\" instruction (w = k - w). Check if previous instruction is a skip instruction. ");
	warn("RETFIE", "SX saves/restores W, STATUS, and FSR automatically.");
	table("ADDLW", "MOV\tHack, W\n\tMOV\tW, ", "ADD\tW, Hack", Literal);
	table("RETFIE", "RETI", NoOperand);
	table("RETURN", "RET", NoOperand);
	//Some of macro instructions (bult-in in MPASM)
	table("CLRC", "CLC", NoOperand);
	table("CLRDC", "CLRB\ttDC", NoOperand);
	table("CLRZ", "CLZ", NoOperand);
	table("MOVFW", "MOV\tW, f", FileOperand);
	table("SETC", "STC", NoOperand);
	table("SETDC", "SETB\tDC", NoOperand);
	table("SETZ", "STZ", NoOperand);
	table("SKPC", "SC", NoOperand);
	table("SKPDC", "SB\tDC", NoOperand);
	table("SKPZ", "SZ", NoOperand);
	table("SKPNC", "SNC", NoOperand);
	table("SKPNDC", "SNB\tDC", NoOperand);
	table("SKPNZ", "SNZ", NoOperand);
	table("TSTF", "TEST\tf", FileOperand);
	table("B", "JMP", Label);
	table("BC", "JC", Label);
	table("BDC", "SNB\tDC\n\tJMP", Label);
	table("BZ", "JZ", Label);
	table("BNC", "JNC", Label);
	table("BNDC", "SB\tDC\n\tJMP", Label);
	table("BNZ", "JNZ", Label);
	table("LCALL", "CALL", LongLabel);
	table("LJMP", "JMP", LongLabel);
	//directives
	table("ORG", "ORG\tf", FileOperand);
	table("EQU", "EQU\tf", FileOperand);
	//table("#define", "EQU\tf", FileOperand);
	//Warn on unsupported MPASM macros
	warn("ADDCF", "MPASM macro ADDCF is not supported yet. Replace manually.");
	warn("ADDDCF", "MPASM macro ADDDCF is not supported yet. Replace manually.");
	warn("NEGF", "MPASM macro NEGF is not supported yet. Replace manually.");
	warn("SUBCF", "MPASM macro SUBCF is not supported yet. Replace manually.");
	warn("SUBDCF", "MPASM macro SUBDCF is not supported yet. Replace manually.");
	//warn("...", "MPASM macro ... is not supported yet. Replace manually.");

	//CBLOCK is a special case
	if(mnemonic == "CBLOCK")
	{
		mnemonic = "\tORG\tf";	//replace by ORG
		operandsType = FileOperand;
		inCblock = true; //switch mode
	}

	//convert mnemonic case if necessary
	if(lowerCase)
	{
	 lowercaseMnemonic(mnemonic);
	 lowercaseMnemonic(mnemonic2);
	}
}

void
MPASMFileLine::translateLineTo(FILE* fout)
{
	//print warning if necessary
	if(!warning.empty())
	{
		fprintf(fout, ";*** WARNING: %s\n", warning.c_str());
		//quote source line if instruction is understood (to preserve source)
		if(operandsType != NotApplicable && !echoSourceLine)
		{
			fprintf(fout, ";");
			printSourceLine(fout);
			fprintf(fout, "\n");
		}
	}
	if(echoSourceLine)
	{
		fprintf(fout, ";");
		printSourceLine(fout);
		fprintf(fout, "\n");
	}
	switch(operandsType)
	{
	case Literal:
		translateLiteralInstruction(fout);
		break;
	case LiteralNoSharp:
		translateLiteralNoSharpInstruction(fout);
		break;
	case FileOperand:
		translateFileOperandInstruction(fout);
		break;
	case FileAndDestination:
		translateFileAndDestinationInstruction(fout);
		break;
	case BitOperand:
		translateBitOperandInstruction(fout);
		break;
	case NoOperand:
		translateNoOperandInstruction(fout);
		break;
	case PortX:
		translatePortX_Instruction(fout);
		break;
	case TwoChoices:
		translateTwoChoicesInstruction(fout);
		break;
	case Label:
		translateLabelInstruction(fout);
		break;
	case LongLabel:
		translateLongLabelInstruction(fout);
		break;
	case NotApplicable:
		translateNotApplicableInstruction(fout);
		break;
	case Directive:
		translateDirectiveInstruction(fout);
		break;
	default:
		break;
	}
	fprintf(fout, "\n");
}

inline
void 
MPASMFileLine::table(const char* x, const char* y, OPTYPE op)
{
	if(mnemonic == x)
	{
		mnemonic = y;
		operandsType = op;
	}
}

inline
void 
MPASMFileLine::table(const char* x, const char* y, const char* y2, OPTYPE op)
{
	if(mnemonic == x)
	{
		mnemonic = y;
		mnemonic2 = y2;
		operandsType = op;
	}
}


void
MPASMFileLine::warn(const char* x, const char* y)
{
	if(mnemonic == x)
	{
		warning += y;
	}
}

void
MPASMFileLine::extractOperands(int s)
{
	string buf;
	int x;
	
	//copy rest of line from s
	buf = sourceLine.substr(s, sourceLine.size() - s);
	//extract comment
	x = 0;
	while(x != -1)
	{
		x = buf.find(';', x);
		//check if this is a part of special charatcters:
		//&amp;&lt;&gt;&quot;
		if(x != -1
			&& buf.substr(x >=4 ? x - 4 : 0, 4) != "&amp"
			&& buf.substr(x >=3 ? x - 3 : 0, 3) != "&lt"
			&& buf.substr(x >=3 ? x - 3 : 0, 3) != "&gt"
			&& buf.substr(x >=5 ? x - 5 : 0, 5) != "&quot")
		{
			//no, it's comment
			comment = buf.substr(x, buf.size() - x);
			stripBlanks(comment);
			buf.resize(x);
		}
					 if(x != -1)
					 {
						 x++;
					 }
	}
	//take a look for one operand or two operands divided by a comma
	//find operand separator ','
	x = buf.find(',');
	if( x != -1)
	{
		//work with second operand
		operand2 = buf.substr(x, buf.size() - x);
		operand2.erase(0, 1); //delete comma
		stripBlanks(operand2);
		buf.resize(x);
	}
	//one operand left
	operand1 = buf;
	stripBlanks(operand1);
	checkOperand(operand1);
	checkOperand(operand2);
	fixStatusFlags();
}

void
MPASMFileLine::stripBlanks(string& buf)
{
	while(!buf.empty() && isspace(buf.at(0)))
	{
		buf.erase(0, 1);
	}
	while(!buf.empty() && isspace(buf.at(buf.size() - 1)))
	{
		buf.erase(buf.size() - 1, 1);
	}
}

void 
MPASMFileLine::translateLiteralInstruction(FILE* fout)
{
	//to add: "\t%s"
	printLabel(fout);
	fprintf(fout, "%s#%s", mnemonic.c_str(), operand1.c_str());
	printComment(fout);
	if(!mnemonic2.empty())
	{
		fprintf(fout, "\n\t%s", mnemonic2.c_str());
	}
}

void 
MPASMFileLine::translateLiteralNoSharpInstruction(FILE* fout)
{
	//to add: "\t%s"
	printLabel(fout);
	fprintf(fout, "%s%s", mnemonic.c_str(), operand1.c_str());
	printComment(fout);
	if(!mnemonic2.empty())
	{
		fprintf(fout, "\n\t%s", mnemonic2.c_str());
	}
}

void 
MPASMFileLine::translateFileOperandInstruction(FILE* fout)
{
	int x;
	string temp;
	//to add: "\t%s" and replace f by actual name
	
	temp = mnemonic;
	x = temp.find('f');
	if(x != -1)
	{
		temp.erase(x, 1);
		temp.insert(x, operand1);
	}
	printLabel(fout);
	fprintf(fout, "%s", temp.c_str());
	printComment(fout);
}

void 
MPASMFileLine::translateFileAndDestinationInstruction(FILE* fout)
{
	//Reg, f >> Reg, W
	//Reg, w >> W, Reg
	printLabel(fout);
	fprintf(fout, "%s\t", mnemonic.c_str());
	if(isDestinationFile())
	{
		fprintf(fout, "%s, W", operand1.c_str());
	}
	else
	{
		fprintf(fout, "W, %s", operand1.c_str());
	}
	printComment(fout);
}

void 
MPASMFileLine::translateBitOperandInstruction(FILE* fout)
{
	printLabel(fout);
	fprintf(fout, "%s\t%s", mnemonic.c_str(), operand1.c_str());
	if(!operand2.empty())
	{
		fprintf(fout, ".%s", operand2.c_str());
	}
	printComment(fout);
}

void 
MPASMFileLine::translateNoOperandInstruction(FILE* fout)
{
	printLabel(fout);
	fprintf(fout, "%s", mnemonic.c_str());
	printComment(fout);
}

void 
MPASMFileLine::translateDirectiveInstruction(FILE* fout)
{
	fprintf(fout, "%s", mnemonic.c_str());
	printComment(fout);
}

void 
MPASMFileLine::translatePortX_Instruction(FILE* fout)
{
	warning += "This instruction conversion is not implemented yet ";
	printComment(fout);
}

void 
MPASMFileLine::translateTwoChoicesInstruction(FILE* fout)
{
	int x;
	string temp;
	//Reg, f - first
	//Reg, w - second
	//In any case replace f by actual name, destination may be skipped (default destination is f)
	
	if(isDestinationFile())
	{
		temp = mnemonic;
	}
	else
	{
		temp = mnemonic2;
	}
	x = temp.find('f');
	if(x != -1)
	{
		temp.replace(x, 1, operand1);
	}
	printLabel(fout);
	fprintf(fout, "%s", temp.c_str());
	printComment(fout);
}

void 
MPASMFileLine::translateLabelInstruction(FILE* fout)
{
	printLabel(fout);
	fprintf(fout, "%s\t%s", mnemonic.c_str(), operand1.c_str());
	printComment(fout);
}

void 
MPASMFileLine::translateLongLabelInstruction(FILE* fout)
{
	printLabel(fout);
	fprintf(fout, "%s\t@%s", mnemonic.c_str(), operand1.c_str());
	printComment(fout);
}

void 
MPASMFileLine::translateNotApplicableInstruction(FILE* fout)
{
	//instruction was not recognized
	//fix constants in source
	if(fixConstantsInComments)
	{
		fixConstantsIn(sourceLine);
	}
	printSourceLine(fout);
}

void
MPASMFileLine::printComment(FILE* fout)
{
	//print a comment if necessary
	if(!comment.empty())
	{
		if(fixConstantsInComments)
		{
			fixConstantsIn(comment);
		}
		fprintf(fout, "\t%s", comment.c_str());
	}
}

void
MPASMFileLine::printLabel(FILE* fout)
{
	string temp = "";
	
	temp = sourceLine.substr(0, mnemonicStart);
	stripBlanks(temp);
	//print a label with a tab
	fprintf(fout, "%s\t", temp.c_str());
}

void
MPASMFileLine::printSourceLine(FILE* fout)
{
	//remove spaces at the end
	while(!sourceLine.empty() && isspace(sourceLine.at(sourceLine.size() - 1)))
	{
		sourceLine.erase(sourceLine.size() - 1, 1);
	}
	fprintf(fout, "%s", sourceLine.c_str());
}

void 
MPASMFileLine::checkOperand(string& op)
{
	int x;
	string copy;
	
	copy = op;
	uppercaseString(copy);
	
	if(copy.find("OPTION") != -1)
	{
		warning += "OPTION register is accessed only by MOV !OPTION, W. ";
	}
	if(copy.find("TRIS") != -1)
	{
		warning += "TRIS registers are accessed by MOV !Rx, W (M = $0F or $1F). ";
	}
	if(copy.find("PCLATH") != -1)
	{
		warning += "PCLATH register bits are in STATUS PAx bits. Or use PAGE/IREAD if possible";
	}
	else
	{
		x = copy.find("PCL"); 
		if(x != -1 && copy.size() == 3)
		{
			op.replace(x, 3, "PC");
		}
	}
	x = copy.find("PORT");
	if(x != -1 && copy.size() == 5)
	{
		op.replace(x, 4, "R");
	}
	x = copy.find("TMR0");
	if(x != -1 && copy.size() == 4)
	{
		op.replace(x, 4, "RTCC");
	}
	fixConstantsIn(op);
}

bool 
MPASMFileLine::isDestinationFile()
{
	string temp;
	
	temp = operand2;
	uppercaseString(temp);
	if(temp == "")
	{
		return(true);
	}
	else if(temp == "W"
		|| temp == "0"
		|| temp == "$0")
	{
		return(false);
	}
	else if(temp == "F"
		|| temp == "1"
		|| temp == "$1")
	{
		return(true);
	}
	warning += "Can't understand destination. Check if it's OK. ";
	return(true);
}

void 
MPASMFileLine::fixBinaryConstantsIn(string& op)
{
	int s = 0;
	int x;
	string copy;
	
	copy.append(op);
	uppercaseString(copy);
	while((s = copy.find("B'", s)) != -1)
	{
		x = s + 1;
		while(++x < op.size())
		{
			if(op.at(x) != '0' && op.at(x) != '1')
			{
				if(op.at(x) == '\'')
				{
					op.erase(x, 1);
					copy.erase(x, 1);
					op.replace(s, 2, "%");
					copy.replace(s, 2, "%");
					x -= 2;
				}
				break;
			}
		}
		s = x;
	}
}

void 
MPASMFileLine::fixDecimalConstantsIn(string& op)
{
	int s = 0;
	int x;
	string copy;
	
	copy.append(op);
	uppercaseString(copy);
	//fix D'123' notation
	while((s = copy.find("D'", s)) != -1)
	{
		x = s + 1;
		while(++x < op.size())
		{
			if(!isdigit(op.at(x)))
			{
				if(op.at(x) == '\'')
				{
					op.erase(x, 1);
					copy.erase(x, 1);
					op.erase(s, 2);
					copy.erase(s, 2);
					x -= 3;
				}
				break;
			}
		}
		s = x;
	}
	//fix .123 notation
	s = 0;
	while((s = copy.find('.', s)) != -1)
	{
		x = s + 1;
		if(x < op.size() && isdigit(op.at(x)))
		{
			op.erase(s, 1);
			copy.erase(s, 1);
			x--;
		}
		s = x;
	}
}

void 
MPASMFileLine::fixHexadecimalConstantsIn(string& op)
{
	int s = 0;
	int x;
	string copy;
	bool wasLetter = false;
	
	copy.append(op);
	uppercaseString(copy);
	//fix H'123ABC' notation
	while((s = copy.find("H'", s)) != -1)
	{
		x = s + 1;
		while(++x < op.size())
		{
			if(!isxdigit(op.at(x)))
			{
				if(op.at(x) == '\'')
				{
					op.erase(x, 1);
					copy.erase(x, 1);
					op.replace(s, 2, "$");
					copy.replace(s, 2, "$");
					x -= 2;
				}
				break;
			}
		}
		s = x;
	}
	//fix 0xAB123 notation
	s = 0;
	while((s = copy.find("0X", s)) != -1)
	{
		x = s + 2;
		if(x < op.size() && isxdigit(op.at(x)))
		{
			op.replace(s, 2, "$");
			copy.replace(s, 2, "$");
			x--;
		}
		s = x;
	}
	//fix if default radix is hex
	if(radix == 16)
	{
		//starts with a decimal digit
		//before starting digit there is no letter or '_'
		s = 0;
		x = -1;
		while(s < copy.size())
		{
			if(isdigit(copy.at(s)))
			{
				//start of number
				if(!wasLetter && x == -1)
				{
					x = s;
				}
			}
			else if(isxdigit(copy.at(s)) && x != -1)
			{
				//continuation of number
			}
			else if(isalpha(copy.at(s))
				|| copy.at(s) == '$'
				|| copy.at(s) == '\''
				|| copy.at(s) == '%'
				|| copy.at(s) == '.')
			{
				x = -1;
				wasLetter = true;
			}
			else
			{
				if(x != -1)
				{
					//number end - insert $
					copy.insert(x, "$");
					op.insert(x, "$");
					s++;
				}
				x = -1;
				wasLetter = false;
			}
			s++;
		}
		if(x != -1)
		{
			//number end - insert $
			op.insert(x, "$");
		}
	}
}

void 
MPASMFileLine::uppercaseString(string& copy)
{
	int x = 0;
	while(x < copy.size())
	{
		copy.at(x) = toupper(copy.at(x));
		x++;
	}
}

void 
MPASMFileLine::lowercaseMnemonic(string& m)
{
	int x = 0;
	while(x < m.size())
	{
		if(isalpha(m.at(x)))
		{
			m.at(x) = tolower(m.at(x));
			x++;
		}
		else
		{
			//check is there another mnemonic
			x = m.find("\n", x);
			if(x == -1)
			{
				break; //no more mnemonics
			}
			x += 2;	//compensate for tabs
			//cause new mnemonic is separated by a new line
			//and a tab
		}
		
	}
}

void 
MPASMFileLine::fixStatusFlags()
{
	string op1;
	string op2;
	
	if(operandsType == BitOperand)
	{
		op1 = operand1;
		op2 = operand2;
		//uppercase copy of operands to check
		//for errors when lower case names are used
		uppercaseString(op1);
		uppercaseString(op2);
		if(op1 == "STATUS")
		{
			if(op2 == "C" || op2 == "CARRY")
			{
				operand1 = "C";
				operand2 = "";
			}
			if(op2 == "DC" || op2 == "DIGITCARRY")
			{
				operand1 = "DC";
				operand2 = "";
			}
			if(op2 == "Z" || op2 == "ZERO")
			{
				operand1 = "Z";
				operand2 = "";
			}
		}
		else
		{
			if(op1 == "_C")
			{
				operand1 = "C";
				operand2 = "";
			}
			if(op1 == "_DC")
			{
				operand1 = "DC";
				operand2 = "";
			}
			if(op1 == "_Z")
			{
				operand1 = "Z";
				operand2 = "";
			}
		}
	}
}

void 
MPASMFileLine::fixConstantsIn(string& op)
{
	fixBinaryConstantsIn(op);
	fixHexadecimalConstantsIn(op);
	fixDecimalConstantsIn(op);      //decimal must go last because
	//it removes any special character before number
}

