#ifndef MPASMFileLine_H
#define MPASMFileLine_H

enum OPTYPE {Literal,
	LiteralNoSharp,
	FileOperand,
	FileAndDestination,
	BitOperand,
	NoOperand,
	PortX,
	TwoChoices,
	Label,
	LongLabel,
	NotApplicable,
	Directive
};

class MPASMFileLine
{
public:
	MPASMFileLine();
	~MPASMFileLine();
	static bool isHTML;
	static bool echoSourceLine;
	static bool lowerCase;
	static char radix;
	static bool fixConstantsInComments;
	void initialize();
	void readSourceLine(FILE* fin);
	void readSourceLine(string& fin, int& start);
	void extractMnemonic();
	void extractCblock();
	void decodeSourceLine();
	void table(const char* x, const char* y, OPTYPE op);
	void table(const char* x, const char* y, const char* y2, OPTYPE op);
	void warn(const char* x, const char* y);
	void extractOperands(int s); //extract operands and comments starting from s in source line
	void stripBlanks(string& buf); //strip leading and trailing spaces in a string
	//File-out functions
	void translateLineTo(FILE* fout);	//main function
		//sub-functions
	void translateLiteralInstruction(FILE* fout);
	void translateLiteralNoSharpInstruction(FILE* fout);
	void translateFileOperandInstruction(FILE* fout);
	void translateFileAndDestinationInstruction(FILE* fout);
	void translateBitOperandInstruction(FILE* fout);
	void translateNoOperandInstruction(FILE* fout);
	void translatePortX_Instruction(FILE* fout);
	void translateTwoChoicesInstruction(FILE* fout);
	void translateLabelInstruction(FILE* fout);
	void translateLongLabelInstruction(FILE* fout);
	void translateNotApplicableInstruction(FILE* fout);
	void translateDirectiveInstruction(FILE* fout);
	void printComment(FILE* fout);
	void printLabel(FILE* fout);
	void printSourceLine(FILE* fout);
	void checkOperand(string& op);	//check if operands contain 'dangerous' regs, e.g. OPTION
	bool isDestinationFile(); //checks operand2 for destination (default is file)
	void fixConstantsIn(string& op);
	void fixBinaryConstantsIn(string& op);
	void fixDecimalConstantsIn(string& op);
	void fixHexadecimalConstantsIn(string& op);
	void uppercaseString(string& copy);
	void lowercaseMnemonic(string& m);
	void fixStatusFlags(); //for bit operands
protected:
private:
	bool inCblock;
	string sourceLine;
	string sourceField; //possibly a mnemonic ( from readSource(fin) )
	int mnemonicStart; //in sourceLine
	string mnemonic;
	string mnemonic2;
	string operand1;
	string operand2;
	string comment;
	string warning;	//if a warning copy commented source line
//	InstructionType type;
	OPTYPE operandsType;
};


#endif